package de.towerdefence.server.game.pathfinding;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.joml.Vector2i;

import lombok.AllArgsConstructor;

@AllArgsConstructor
public class AStar {
    private final WalkableCallback walkable;

    public Vector2i next(Vector2i current, Vector2i target) throws NoPathException, OutOfBoundsException {
        if (!walkable.call(current) || !walkable.call(target)) {
            throw new OutOfBoundsException();
        }

        if (current.equals(target)) {
            return target;
        }

        Map<Vector2i, PathNode> remaining = new HashMap<>();
        Set<Vector2i> evaluated = new HashSet<>();
        Map<Vector2i, Vector2i> takenPath = new HashMap<>();

        PathNode origin = new PathNode(current, 0, current.gridDistance(target));
        remaining.put(current, origin);

        while (!remaining.isEmpty()) {
            PathNode currentNode = getLowestEstimate(remaining);
            Vector2i currentPos = currentNode.pos;

            if (currentPos.equals(target)) {
                return reconstructPath(takenPath, current);
            }

            remaining.remove(currentPos);
            evaluated.add(currentPos);

            long nextCost = currentNode.cost + 1;
            for (PathNode neighbor : getNeighbors(currentPos, nextCost, target)) {
                if (evaluated.contains(neighbor.pos)) {
                    continue;
                }
                if (!remaining.containsKey(neighbor.pos) || nextCost < remaining.get(neighbor.pos).cost) {
                    takenPath.put(neighbor.pos, currentPos);
                    remaining.put(neighbor.pos, neighbor);
                }
            }
        }

        throw new NoPathException();
    }

    private Vector2i reconstructPath(Map<Vector2i, Vector2i> takenPath, Vector2i origin) {
        Vector2i current = takenPath.keySet().stream().reduce((first, second) -> second).orElse(null);
        if (current == null)
            return null;
        while (takenPath.get(current) != null && !takenPath.get(current).equals(origin)) {
            current = takenPath.get(current);
        }
        return current;
    }

    private PathNode getLowestEstimate(Map<Vector2i, PathNode> openSet) {
        PathNode lowestNode = null;
        for (PathNode node : openSet.values()) {
            if (lowestNode == null || node.estimate < lowestNode.estimate) {
                lowestNode = node;
            }
        }
        return lowestNode;
    }

    private List<PathNode> getNeighbors(Vector2i pos, long nextCost, Vector2i target) {
        List<PathNode> neighbors = new ArrayList<>();

        addNeighbor(neighbors, new Vector2i(pos.x + 1, pos.y), nextCost, target);
        addNeighbor(neighbors, new Vector2i(pos.x - 1, pos.y), nextCost, target);
        addNeighbor(neighbors, new Vector2i(pos.x, pos.y + 1), nextCost, target);
        addNeighbor(neighbors, new Vector2i(pos.x, pos.y - 1), nextCost, target);
        return neighbors;
    }

    private void addNeighbor(List<PathNode> accumulator, Vector2i pos, long nextCost, Vector2i target) {
        Vector2i right = new Vector2i(pos.x + 1, pos.y);
        if (!walkable.call(pos)) {
            return;
        }
        accumulator.add(
            new PathNode(
                right,
                nextCost,
                right.gridDistance(target) + nextCost
            )
        );
    }
}
