package de.towerdefence.server.server.channels.match;

import com.fasterxml.jackson.databind.ObjectMapper;
import de.towerdefence.server.game.Enemy;
import de.towerdefence.server.game.exeptions.InvalidPlacementException;
import de.towerdefence.server.match.MatchService;
import de.towerdefence.server.match.NotInMatchException;
import de.towerdefence.server.player.Player;
import de.towerdefence.server.server.JsonWebsocketHandler;
import de.towerdefence.server.server.channels.match.enemy.EnemyUpdateMessage;
import de.towerdefence.server.server.channels.match.hitpoints.PlayerHitpointsMessage;
import de.towerdefence.server.server.channels.match.money.PlayerMoneyMessage;
import de.towerdefence.server.server.channels.match.placing.GamePlayerMap;
import de.towerdefence.server.server.channels.match.placing.InvalidPlacementMessage;
import de.towerdefence.server.server.channels.match.placing.RequestTowerPlacingMessage;
import de.towerdefence.server.server.channels.match.placing.TowerPlacedMessage;
import de.towerdefence.server.session.Channel;
import de.towerdefence.server.session.SessionsService;
import java.io.IOException;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import org.springframework.web.socket.CloseStatus;
import org.springframework.web.socket.TextMessage;
import org.springframework.web.socket.WebSocketSession;

public class MatchWebsocketHandler extends JsonWebsocketHandler {

    protected final Map<Player, WebSocketSession> playerSessions = new ConcurrentHashMap<>();
    private final MatchService matchService;
    private final ObjectMapper objectMapper = new ObjectMapper();

    public MatchWebsocketHandler(SessionsService sessionsService, MatchService matchService) {
        super(Channel.MATCH, sessionsService);
        this.matchService = matchService;
    }

    @Override
    public void afterConnectionEstablished(WebSocketSession session) {
        super.afterConnectionEstablished(session);
        playerSessions.put(sessionPlayers.get(session), session);
        matchService.playerConnected(
            sessionPlayers.get(session),
            this::onPlayerMoneyChanged,
            this::onPlayerHitpointsChanged,
            this::onEnemiesChanged
        );
    }

    @Override
    protected void closeSession(WebSocketSession session, CloseStatus reason) {
        Player player = sessionPlayers.get(session);
        super.closeSession(session, reason);
        playerSessions.remove(player);
    }

    private void onPlayerMoneyChanged(Player player, int playerMoney) {
        WebSocketSession session = playerSessions.get(player);
        try {
            new PlayerMoneyMessage(playerMoney).send(session);
        } catch (IOException ignored) {
        }
    }

    private void onPlayerHitpointsChanged(Player player, int playerHitpoints) {
        WebSocketSession session = playerSessions.get(player);
        try {
            new PlayerHitpointsMessage(playerHitpoints).send(session);
        } catch (IOException ignored) {
        }
    }

    private void onEnemiesChanged(Player player, Enemy[] newEnemies, Enemy[] changedEnemies) {
        WebSocketSession session = playerSessions.get(player);
        try {
            new EnemyUpdateMessage(newEnemies, changedEnemies).send(session);
        } catch (IOException ignored) {
        }
    }

    @Override
    protected void handleTextMessage(WebSocketSession session, TextMessage message) {
        try {
            String payload = message.getPayload();
            if (!Objects.equals(
                        objectMapper.readTree(payload).get("$id").asText(),
                        RequestTowerPlacingMessage.MESSAGE_ID
                    )) {
                this.closeSession(session, CloseStatus.BAD_DATA);
                return;
            }
            handleRequestTowerPlacingMessage(session, payload);
        } catch (IOException ignored) {
            this.closeSession(session, CloseStatus.BAD_DATA);
        }
    }

    private void handleRequestTowerPlacingMessage(WebSocketSession session, String payload)
    throws IOException {
        RequestTowerPlacingMessage msg =
        objectMapper.readValue(payload, RequestTowerPlacingMessage.class);
        Player opponent;
        try {
            opponent =this.matchService.placeTower(
                this.sessionPlayers.get(session),
                msg.getX(),
                msg.getY()
            );
        } catch (InvalidPlacementException exception) {
            new InvalidPlacementMessage(msg.getX(), msg.getY(), exception.getReason()).send(session);
            return;
        } catch (NotInMatchException ignored) {
            this.closeSession(session, CloseStatus.BAD_DATA);
            return;
        }
        WebSocketSession opponentSession = playerSessions.get(opponent);
        new TowerPlacedMessage(msg.getX(), msg.getY(), GamePlayerMap.PLAYER).send(session);
        new TowerPlacedMessage(msg.getX(), msg.getY(), GamePlayerMap.OPPONENT).send(opponentSession);
    }
}
